#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from unittest import TestLoader, TextTestRunner, TestSuite

from pycocotb.tests.verilatorCntr_test import VerilatorCntrTC
from pycocotb.tests.verilatorHierarchy_test import VerilatorHierarchyTC
from pycocotb.tests.wire_test import VerilatorWireTC
from pycocotb.tests.verilatorHandshakedWire_test import VerilatorHandshakedWireTC
from pycocotb.tests.i2c_test import I2cAgent_TC


def testSuiteFromTCs(*tcs):
    loader = TestLoader()
    loadedTcs = [loader.loadTestsFromTestCase(tc) for tc in tcs]
    suite = TestSuite(loadedTcs)
    return suite


suite = testSuiteFromTCs(
    # basic tests
    VerilatorCntrTC,
    VerilatorWireTC,
    VerilatorHierarchyTC,
    VerilatorHandshakedWireTC,
    #I2cAgent_TC,
)

if __name__ == '__main__':
    runner = TextTestRunner(verbosity=2)
    runner.run(suite)



from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.process_utils import OnRisingCallbackLoop
from pycocotb.triggers import Timer, WaitWriteOnly, WaitCombRead


def get_clk_driver(sim: HdlSimulator, clk, clk_period):
    while True:
        yield WaitWriteOnly()
        clk.write(0)

        yield Timer(clk_period // 2)
        yield WaitWriteOnly()

        clk.write(1)
        yield Timer(clk_period // 2)


def get_rst_driver(sim: HdlSimulator, rst, delay):
    yield WaitWriteOnly()
    assert sim.now == 0
    rst.write(1)

    yield Timer(delay)
    assert sim.now == delay
    yield WaitWriteOnly()
    assert sim.now == delay
    rst.write(0)


def get_pull_up_driver(sim: HdlSimulator, sig, delay):
    yield WaitWriteOnly()
    sig.write(0)

    yield Timer(delay)
    assert sim.now == delay
    yield WaitWriteOnly()
    sig.write(1)


def get_pull_up_driver_with_reset(sim: HdlSimulator, sig, reset, clk_period):
    exp_t = 0
    yield WaitWriteOnly()
    sig.write(0)
    assert sim.now == exp_t

    while True:
        yield WaitCombRead()
        if not reset.read():
            assert sim.now == exp_t
            yield WaitWriteOnly()
            sig.write(1)
            return
        else:
            yield Timer(clk_period)
            exp_t += clk_period


def get_sync_pull_up_driver_with_reset(sim: HdlSimulator, sig, clk, rst):
    # init
    yield WaitWriteOnly()
    sig.write(0)
    assert sim.now == 0

    def pull_up_after():
        exp_t = sim.now
        yield WaitCombRead()
        assert sim.now == exp_t

        if not rst.read():
            yield WaitWriteOnly()
            sig.write(1)
            assert sim.now == exp_t

    yield OnRisingCallbackLoop(sim, clk, pull_up_after, lambda: True)()


def get_sync_sig_monitor(sim: HdlSimulator, sig, clk, rst, result):

    def monitorWithClk():
        # if clock is specified this function is periodically called every
        # clk tick
        yield WaitCombRead()
        if not rst.read():
            result.append((sim.now, int(sig.read())))

    return OnRisingCallbackLoop(sim, clk, monitorWithClk, lambda: True)()




from typing import Tuple

from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.process_utils import OnRisingCallbackLoop

# The constant which means that the agent shouLd wait one time quantum
# before sending a new data over an interface.
NOP = "NOP"
RX = "RX"
TX = "TX"


class AgentBase():
    """
    Base class of agent of interface like in UVM
    driver is used for slave interfaces
    monitor is used for master interfaces

    :ivar ~.intf: interface assigned to this agent
    :ivar ~._enable: flag to enable/disable this agent
    :ivar ~._debugOutput: optional stream where to print debug messages
    """
    # because ohterwise there will be a cycle and python
    # will not be able to deallocate this and sim/intf
    __weakref__ = ["intf", "sim"]

    def __init__(self, sim: HdlSimulator, intf):
        self.intf = intf
        self._enabled = True
        self._debugOutput = None
        self.sim = sim

    def setEnable(self, en):
        self._enabled = en

    def getEnable(self):
        return self._enabled

    def _debug(self, out):
        self._debugOutput = out

    def getDrivers(self):
        """
        Called before simulation to collect all drivers of interfaces
        from this agent
        """
        return [self.driver(), ]

    def getMonitors(self):
        """
        Called before simulation to collect all monitors of interfaces
        from this agent
        """
        return [self.monitor(), ]

    def driver(self):
        """
        Implement this method to drive your interface
        in simulation/verification
        """
        raise NotImplementedError()

    def monitor(self):
        """
        Implement this method to monitor your interface
        in simulation/verification
        """
        raise NotImplementedError()


class AgentWitReset(AgentBase):

    def __init__(self, sim: HdlSimulator, intf, rst: Tuple["RtlSignal", bool]):
        """
        :param rst: tuple (rst signal, rst_negated flag)
        """
        super(AgentWitReset, self).__init__(sim, intf)
        rst, rst_negated = rst
        self.rst = rst
        self.rstOffIn = int(rst_negated)

    def notReset(self):
        if self.rst is None:
            return True

        rstVal = self.rst.read()
        rstVal = int(rstVal)
        return rstVal == self.rstOffIn


class SyncAgentBase(AgentWitReset):
    """
    Agent which runs only monitor/driver function at specified edge of clk
    """
    SELECTED_EDGE_CALLBACK = OnRisingCallbackLoop

    def __init__(self, sim: HdlSimulator,
                 intf,
                 clk: "RtlSignal",
                 rst: Tuple["RtlSignal", bool],
                  wrap_monitor_and_driver_in_edge_callback=True):
        super(SyncAgentBase, self).__init__(
            sim, intf, rst)
        self.clk = clk

        if wrap_monitor_and_driver_in_edge_callback:
            # run monitor, driver only on rising edge of clk
            c = self.SELECTED_EDGE_CALLBACK
            self.monitor = c(sim, self.clk, self.monitor, self.getEnable)
            self.driver = c(sim, self.clk, self.driver, self.getEnable)

    def setEnable_asDriver(self, en: bool):
        self._enabled = en
        self.driver.setEnable(en)

    def setEnable_asMonitor(self, en: bool):
        self._enabled = en
        self.monitor.setEnable(en)

    def getDrivers(self):
        self.setEnable = self.setEnable_asDriver
        return AgentBase.getDrivers(self)

    def getMonitors(self):
        self.setEnable = self.setEnable_asMonitor
        return AgentBase.getMonitors(self)




from tempfile import TemporaryDirectory
import unittest

from pycocotb.constants import CLK_PERIOD
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.tests.common import build_sim
from pycocotb.triggers import WaitCombStable


class VerilatorHierarchyTC(unittest.TestCase):
    """
    Simple test of verilator simulation wrapper access on multiple hierarchy levels
    """

    def build_handshaked_fifo(self, build_dir):
        DATA_WIDTH = 8
        accessible_signals = [
            # (signal_name, read_only, is_signed, type_width)
            ("clk", 0, 0, 1),
            ("dataIn_data", 0, 0, DATA_WIDTH),
            ("dataIn_rd", 1, 0, 1),
            ("dataIn_vld", 0, 0, 1),
            ("dataOut_data", 1, 0, DATA_WIDTH),
            ("dataOut_rd", 0, 0, 1),
            ("dataOut_vld", 1, 0, 1),
            ("rst_n", 0, 0, 1),
            ("size", 1, 0, 3),

            # (("fifo_inst", "clk"), 1, 0, 1),
            # (("fifo_inst", "dataIn_data"), 1, 0, DATA_WIDTH),
            # (("fifo_inst", "dataIn_wait"), 1, 0, 1),
            # (("fifo_inst", "dataIn_en"), 1, 0, 1),
            # (("fifo_inst", "dataOut_data"), 1, 0, DATA_WIDTH),
            # (("fifo_inst", "dataOut_wait"), 1, 0, 1),
            # (("fifo_inst", "dataOut_en"), 1, 0, 1),
            # (("fifo_inst", "rst_n"), 0, 1, 1),
            # (("fifo_inst", "size"), 1, 1, 3),
            (("fifo_inst", "memory"), 1, 0, [3, DATA_WIDTH]),
            (("fifo_inst", "fifo_read"), 1, 0, 1),
            (("fifo_inst", "fifo_write"), 1, 0, 1),
        ]
        files = ["fifo.v", "HandshakedFifo.v"]
        return build_sim(files, accessible_signals, self, build_dir, "HandshakedFifo")

    def test_sim_HandshakedFifo(self):
        # build_dir = "tmp"
        # if True:
        with TemporaryDirectory() as build_dir:
            rtl_sim = self.build_handshaked_fifo(build_dir)
            io = rtl_sim.io
            sim = HdlSimulator(rtl_sim)

            def check_if_can_read():
                yield WaitCombStable()
                assert(len(io.fifo_inst.memory) == 3)
                item0 = io.fifo_inst.memory[0]
                item0.read()
                for i in io.fifo_inst.memory:
                    i.read()

            sim.run(int(CLK_PERIOD * 10.5),
                    extraProcesses=[
                        check_if_can_read()
                    ]
            )


if __name__ == "__main__":
    suite = unittest.TestSuite()
    # suite.addTest(VerilatorWireTC('test_wire64'))
    suite.addTest(unittest.makeSuite(VerilatorHierarchyTC))
    runner = unittest.TextTestRunner(verbosity=3)
    runner.run(suite)





from copy import deepcopy
from distutils.sysconfig import get_config_var
from importlib import machinery
from jinja2.environment import Environment
from jinja2.loaders import PackageLoader
import os
from os.path import dirname
import  platform
from setuptools import Extension
from setuptools.command.build_ext import build_ext
from setuptools.dist import Distribution
from subprocess import check_call
import sys
from typing import List, Dict, Tuple

from pycocotb.verilator.fs_utils import find_files, working_directory


VER_SIM_GEN_BASE = os.path.dirname(__file__)
COCOPY_SRC_DIR = os.path.join(VER_SIM_GEN_BASE, "c_files")
VERILATOR_ROOT = "/usr/local/share/verilator"
VERILATOR_INCLUDE_DIR = os.path.join(VERILATOR_ROOT, "include")
VERILATOR = "verilator_bin_dbg"

template_env = Environment(
    loader=PackageLoader("pycocotb", "verilator/templates")
)
verilator_sim_wrapper_template = template_env.get_template(
    'verilator_sim.cpp.template')

# "cpython-36m-x86_64-linux-gnu"
SOABI = get_config_var("SOABI")
# e.g. "linux"
MACHDEP = get_config_var("MACHDEP")
# e.g. "x86_64"
AR = platform.processor()
# e.g. "3.6"
VERSION = get_config_var("VERSION")

IN_PLACE_LIB_DIR = os.path.abspath(
    os.path.join(
        VER_SIM_GEN_BASE,
        "..",
        "..",
        "build",
        "lib.%s-%s-%s" % (MACHDEP, AR, VERSION),
        "pycocotb",
        "verilator"))
INSTALLED_LIB_DIR = os.path.join(VER_SIM_GEN_BASE)


DEFAULT_EXTENSION_EXTRA_ARGS = {
    "extra_compile_args": ['-std=c++17'],
    "libraries": ["boost_coroutine", "boost_context", "boost_system",
                  "common." + SOABI],
    "include_dirs": [VERILATOR_INCLUDE_DIR, COCOPY_SRC_DIR],
    "language": "c++",
    "library_dirs": [IN_PLACE_LIB_DIR, INSTALLED_LIB_DIR],
    "extra_compile_args": [
        "-std=c++11",
        # "-faligned-new"
    ],
}


def verilatorCompile(files: List[str], build_dir: str):
    include_dirs = ["-I%s" % dn
                    for dn in set(dirname(f)
                                  for f in files)
                    if dn and dn != "."]
    files = [files[-1], ]
    cmd = [VERILATOR, "--cc", "--event-triggers",
           "--trace", "--Mdir", build_dir] + files + include_dirs
    try:
        check_call(cmd)
    except Exception:
        print(" ".join(cmd), file=sys.stderr)
        raise


def getSrcFiles(build_dir: str):
    build_sources = find_files(build_dir, pattern="*.cpp", recursive=True)
    return [*build_sources]


def generatePythonModuleWrapper(
        top_name: str, top_unique_name: str,
        build_dir: str,
        accessible_signals: List[Tuple[str, bool, bool, int]],
        extra_Extension_args: Dict[str, object]=DEFAULT_EXTENSION_EXTRA_ARGS):
    """
    Collect all c/c++ files into setuptools.Extension and build it

    :param top_name: name of top in simulation
    :param top_unique_name: unique name used as name for simulator module
    :param build_dir: tmp directory where simulation should be build
    :param verilator_include_dir: include directory of Verilator
    :param accessible_signals: List of tuples (signal_name, signal_phy_name, read_only, is_signed, type_width)
    :param extra_Extension_args: additional values for setuptools.Extension constructor

    :return: file name of builded module (.so/.dll file)
    """
    with working_directory(build_dir):
        with open("V" + top_name + "_sim_wrapper.cpp", "w") as f:
            f.write(verilator_sim_wrapper_template.render(
                module_name=top_unique_name,
                top_name=top_name,
                accessible_signals=accessible_signals))
        sources = getSrcFiles(".")
        dist = Distribution()

        dist.parse_config_files()

        extra_Extension_args = deepcopy(extra_Extension_args)
        extra_Extension_args["sources"] = extra_Extension_args.get("sources", []) + sources
        extra_Extension_args["include_dirs"] = extra_Extension_args.get("include_dirs", []) + [build_dir, ]

        sim = Extension(top_unique_name,
                        **extra_Extension_args,
                        )

        dist.ext_modules = [sim]
        _build_ext = build_ext(dist)
        _build_ext.finalize_options()
        _build_ext.run()
        return os.path.join(build_dir, _build_ext.build_lib,
                            sim._file_name)


def loadPythonCExtensionFromFile(library_file_name: str, module_name: str):
    importer = machinery.FileFinder(
        os.path.dirname(os.path.abspath(library_file_name)),
        (machinery.ExtensionFileLoader,
         machinery.EXTENSION_SUFFIXES))
    module = importer.find_module(module_name).load_module(module_name)
    return module



from typing import Tuple
from itertools import islice


def sim_eval_cond(v):
    """
    Evaluate list of values as condition

    :return: tuple (value, value valid)
    """
    if v.vld_mask == 1:
        return bool(v.val), True
    else:
        return False, False


def valueHasChanged(valA: "Value", valB: "Value"):
    return valA.val is not valB.val or valA.vld_mask != valB.vld_mask


class ValueUpdater():

    def __init__(self, nextVal, invalidate: bool):
        """
        Create value updater for simulation
    
        :param nextVal: instance of Value which will be assigned to signal
        :param invalidate: flag which tells if value has been compromised
            and if it should be invalidated
        """
        self.nextVal = nextVal
        self.invalidate = invalidate

    def __call__(self, currentVal: "Value"):
        _nextVal = self.nextVal.__copy__()
        if self.invalidate:
            _nextVal.vld_mask = 0
        return (valueHasChanged(currentVal, _nextVal), _nextVal)


class ArrayValueUpdater():
    
    def __init__(self, nextItemVal: "Value", indexes: Tuple["Value"],
                 invalidate: bool):
        """
        Create value updater for simulation for value of array type
        
        :param nextVal: instance of Value which will be assigned to signal
        :param indexes: tuple on indexes where value should be updated
            in target array
        """
        self.nextItemVal = nextItemVal
        self.indexes = indexes
        self.invalidate = invalidate

    def __call__(self, currentVal):
        _currentVal = currentVal
        indexes = self.indexes
        if len(indexes) > 1:
            raise NotImplementedError("[TODO] implement for more indexes")
            for i in islice(indexes, len(indexes) - 1):
                _currentVal = _currentVal[i]

        nextItemVal = self.nextItemVal
        _nextItemVal = nextItemVal.__copy__()
        if self.invalidate:
            _nextItemVal.vld_mask = 0

        index = indexes[-1]
        change = valueHasChanged(_currentVal[index], _nextItemVal)
        _currentVal[index] = _nextItemVal
        return (change, currentVal)



from collections import deque
from typing import Tuple

from pycocotb.agents.base import NOP, SyncAgentBase
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.triggers import WaitCombStable, WaitWriteOnly, WaitCombRead


class HandshakedAgent(SyncAgentBase):
    """
    Simulation/verification agent for handshaked interfaces
    interface there is onMonitorReady(simulator)
    and onDriverWriteAck(simulator) unimplemented method
    which can be used for interfaces with bi-directional data streams

    :note: 2-phase (xor) handshake
    """

    def __init__(self, sim: HdlSimulator, intf, clk: "RtlSignal",
                 rst: Tuple["RtlSignal", bool]):
        super(HandshakedAgent, self).__init__(
            sim, intf, clk, rst)
        self.actualData = NOP
        self.data = deque()

        # tmp variables to keep track of last send values to simulation
        self._lastWritten = None
        self._lastRd = None
        self._lastVld = None
        self._readyComnsummed = True
        # callbacks
        self._afterRead = None
        

    def setEnable_asDriver(self, en):
        super(HandshakedAgent, self).setEnable_asDriver(en)
        if en:
            # pop new data if there are not any pending
            if self.actualData is NOP and self.data:
                self.actualData = self.data.popleft()

            doSend = self.actualData is not NOP
            if doSend:
                self.set_data(self.actualData)
            doSend = int(doSend)
            self.set_valid(doSend)
            self._lastVld = doSend
        else:
            self.set_valid(0)
            self.set_data(None)
            self._lastVld = 0

    def setEnable_asMonitor(self, en):
        super(HandshakedAgent, self).setEnable_asMonitor(en)
        self.set_ready(int(en))
        self._lastRd = int(en)

    def get_ready(self) -> bool:
        """
        get value of "ready" signal
        """
        raise NotImplementedError("Implement this method to read ready signal on your interface")

    def set_ready(self, val: bool):
        raise NotImplementedError("Implement this method to write ready signal on your interface")

    def get_valid(self):
        """
        get value of "valid" signal, override f.e. when you
        need to use signal with reversed polarity
        """
        raise NotImplementedError("Implement this method to read valid signal on your interface")

    def set_valid(self, val):
        raise NotImplementedError("Implement this method to write valid signal on your interface")

    def get_data(self):
        """extract data from interface"""
        raise NotImplementedError("Implement this method to read data signals on your interface")

    def set_data(self, data):
        """write data to interface"""
        raise NotImplementedError("Implement this method to write data signals on your interface")

    def monitor(self):
        """
        Collect data from interface
        If onMonitorReady is present run it before setting ready and before data is read from the channel
        """
        start = self.sim.now
        yield WaitCombRead()
        if not self._enabled:
            return

        if self.notReset():
            yield WaitWriteOnly()
            if not self._enabled:
                return

            if self._readyComnsummed:
                # try to run onMonitorReady if there is any to preset value on signals potentially
                # going against main data flow of this channel
                onMonitorReady = getattr(self, "onMonitorReady", None)
                if onMonitorReady is not None:
                    onMonitorReady()
                self._readyComnsummed  = False

            # update rd signal only if required
            if self._lastRd != 1:
                self.set_ready(1)
                self._lastRd = 1
            else:
                yield WaitCombRead()
                assert int(self.get_ready()) == self._lastRd, (
                    "Something changed the value of ready without notifying this agent"
                    " which is responsible for this",
                    self.sim.now, self.get_ready(), self._lastRd)

            if not self._enabled:
                return

            # wait for response of master
            yield WaitCombStable()
            if not self._enabled:
                return
            vld = self.get_valid()
            try:
                vld = int(vld)
            except ValueError:
                raise AssertionError(
                    self.sim.now, self.intf,
                    "vld signal is in invalid state")

            if vld:
                # master responded with positive ack, do read data
                d = self.get_data()
                if self._debugOutput is not None:
                    self._debugOutput.write(
                        "%s, read, %d: %r\n" % (
                            self.intf._getFullName(),
                            self.sim.now, d))
                self.data.append(d)
                if self._afterRead is not None:
                    self._afterRead()

                # data was read from th channel next ready bellongs to a different data chunk
                self._readyComnsummed = True
        else:
            self._readyComnsummed = True
            if self._lastRd != 0:
                yield WaitWriteOnly()
                # can not receive, say it to masters
                self.set_ready(0)
                self._lastRd = 0
            else:
                assert int(self.get_ready()) == self._lastRd

        assert start == self.sim.now

    def checkIfRdWillBeValid(self):
        yield WaitCombStable()
        rd = self.get_ready()
        try:
            rd = int(rd)
        except ValueError:
            raise AssertionError(self.sim.now, self.intf, "rd signal in invalid state")

    def driver(self):
        """
        Push data to interface

        set vld high and wait on rd in high then pass new data
        """
        start = self.sim.now
        yield WaitWriteOnly()
        if not self._enabled:
            return
        # pop new data if there are not any pending
        if self.actualData is NOP and self.data:
            self.actualData = self.data.popleft()

        doSend = self.actualData is not NOP

        # update data on signals if is required
        if self.actualData is not self._lastWritten:
            if doSend:
                data = self.actualData
            else:
                data = None

            self.set_data(data)
            self._lastWritten = self.actualData

        yield WaitCombRead()
        if not self._enabled:
            return
        en = self.notReset()
        vld = int(en and doSend)
        if self._lastVld is not vld:
            yield WaitWriteOnly()
            self.set_valid(vld)
            self._lastVld = vld

        if not self._enabled:
            # we can not check rd it in this function because we can not wait
            # because we can be reactivated in this same time
            yield self.checkIfRdWillBeValid()
            return

        # wait for response of slave
        yield WaitCombStable()
        if not self._enabled:
            return
        rd = self.get_ready()
        try:
            rd = int(rd)
        except ValueError:
            raise AssertionError(
                self.sim.now, self.intf,
                "rd signal in invalid state") from None

        if not vld:
            assert start == self.sim.now
            return

        if rd:
            # slave did read data, take new one
            if self._debugOutput is not None:
                self._debugOutput.write("%s, wrote, %d: %r\n" % (
                    self.intf._getFullName(),
                    self.sim.now,
                    self.actualData))

            a = self.actualData
            # pop new data, because actual was read by slave
            if self.data:
                self.actualData = self.data.popleft()
            else:
                self.actualData = NOP

            # try to run onDriverWriteAck if there is any
            onDriverWriteAck = getattr(self, "onDriverWriteAck", None)
            if onDriverWriteAck is not None:
                onDriverWriteAck()

            onDone = getattr(a, "onDone", None)
            if onDone is not None:
                onDone()

        assert start == self.sim.now



from setuptools.py31compat import TemporaryDirectory
import unittest

from pycocotb.agents.peripheral.i2c import I2cAgent
from pycocotb.constants import CLK_PERIOD
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.tests.common import build_sim


class Tristate():
    """
    Utility which allows access to tri-state signals by Python object property
    """

    def __init__(self, io, name):
        self.i = getattr(io, name + "_i")
        self.o = getattr(io, name + "_o")
        self.t = getattr(io, name + "_t")

    def as_tuple(self):
        return (self.i, self.o, self.t)


class I2C():
    """
    Utility which allows access to I2C signals by Python object property
    """

    def __init__(self, io, name):
        self.scl = Tristate(io, name + "_scl")
        self.sda = Tristate(io, name + "_sda")

    def as_tuple(self):
        return (self.scl.as_tuple(), self.sda.as_tuple())


class I2cAgent_TC(unittest.TestCase):

    def build_I2c_wire(self, build_dir):
        accessible_signals = [
            # (signal_name, read_only, is_signed, type_width)
            ("i_scl_i", 1, 0, 1),
            ("i_scl_o", 0, 0, 1),
            ("i_scl_t", 0, 0, 1),

            ("i_sda_i", 1, 0, 1),
            ("i_sda_o", 0, 0, 1),
            ("i_sda_t", 0, 0, 1),

            ("o_scl_i", 0, 0, 1),
            ("o_scl_o", 1, 0, 1),
            ("o_scl_t", 1, 0, 1),

            ("o_sda_i", 0, 0, 1),
            ("o_sda_o", 1, 0, 1),
            ("o_sda_t", 1, 0, 1),
        ]
        verilog_files = ["I2c_wire.v"]
        sim = build_sim(verilog_files, accessible_signals, self, build_dir, "I2c_wire")
        i, o = I2C(sim.io, "i"), I2C(sim.io, "o")
        return sim, i, o

    def build_I2c_wire_with_agents(self, build_dir):
        rtl_sim, i, o = self.build_I2c_wire(build_dir)
        sim = HdlSimulator(rtl_sim)

        i_ag = I2cAgent(sim, i.as_tuple(), (None, False))
        o_ag = I2cAgent(sim, o.as_tuple(), (None, False))
        procs = [*i_ag.getDrivers(), *o_ag.getMonitors()]
        # because the pullup is already on other side of interface
        o_ag.sda.pullMode = None
        o_ag.scl.pullMode = None

        return sim, i_ag, o_ag, procs

    def test_nop(self):
        # build_dir = "tmp"
        # if True:
        with TemporaryDirectory() as build_dir:
            sim, i_ag, o_ag, procs = self.build_I2c_wire_with_agents(build_dir)
            ref = [0]
            i_ag.bits.extend(ref)

            # sim.rtl_simulator.set_trace_file("I2c_wire.vcd", -1)
            sim.run(10 * CLK_PERIOD, extraProcesses=procs)
            self.assertSequenceEqual(o_ag.bits,
                                     [I2cAgent.START] + [0 for _ in range(10)])

    def test_simple(self):
        # build_dir = "tmp"
        # if True:
        with TemporaryDirectory() as build_dir:
            sim, i_ag, o_ag, procs = self.build_I2c_wire_with_agents(build_dir)

            ref = [1, 0, 0, 1, 1, 1, 0, 1, 0]
            expected = [I2cAgent.START, 1, 0, 0, 1, 1, 1, 0, 1, 0, 0]
            i_ag.bits.extend(ref)

            # sim.rtl_simulator.set_trace_file("I2c_wire.vcd", -1)
            sim.run(10 * CLK_PERIOD, extraProcesses=procs)
            self.assertSequenceEqual(o_ag.bits,
                                     expected)


if __name__ == "__main__":
    suite = unittest.TestSuite()
    #suite.addTest(I2cAgent_TC('test_simple'))
    suite.addTest(unittest.makeSuite(I2cAgent_TC))
    runner = unittest.TextTestRunner(verbosity=3)
    runner.run(suite)



from sortedcontainers.sorteddict import SortedDict
from typing import Tuple


class DONE:
    pass


class SimTimeSlot():
    """
    :note: write/read only is related
           to access to circuit from python code
    :note: event types
        * timeslot_begin
        * write_only     - w
        * "circuit eval() memory update detected"
        * comb_read      - r
        * comb_stable    - r
        * mem_stable     - r ("rest of circuit eval()")
        * timeslot_end   - r
    :note: if write_only is required in comb_stable or later error is rised
        if it called before the SimTimeSlot is evaluated from beginning
    """
    __slots__ = ['timeslot_begin', 'write_only', 'comb_read', 'comb_stable', 'mem_stable', 'timeslot_end']

    def __init__(self):
        self.timeslot_begin = None
        self.write_only = None
        self.comb_read = None
        self.comb_stable = None
        self.mem_stable = None
        self.timeslot_end = None

    def get_state_name(self):
        if self.timeslot_begin is not DONE:
            return "timeslot_begin"
        elif self.write_only is not DONE:
            return "write_only"
        elif self.comb_read is not DONE:
            return "comb_read"
        elif self.mem_stable is not DONE:
            return "mem_stable"
        elif self.timeslot_end is not DONE:
            return "timeslot_end"
        else:
            return "after timeslot_end"

    def __repr__(self):
        return "<%s in %s>" % (self.__class__.__name__, self.get_state_name())


# internal
class SimCalendar(SortedDict):
    """
    Priority queue where key is time and priority
    """

    def push(self, time: int, value: SimTimeSlot):
        assert isinstance(time, int)
        super(SimCalendar, self).__setitem__(time, value)

    def pop(self) -> Tuple[int, object]:
        return super(SimCalendar, self).popitem(0)



from pycocotb.basic_hdl_simulator.io import BasicRtlSimIo


class BasicRtlSimModel(object):
    """
    Base class for model in simulator
    """

    def __init__(self, sim, name=None):
        self.sim = sim
        self._name = name
        self.io = BasicRtlSimIo()
        self._interfaces = []
        self._processes = []
        self._units = []
        self._outputs = {}

    def _init_body(self):
        """
        This method is used to initializes all containers on this object
        after signals are connected
        """
        pass



from pyMathBitPrecise.array3t import Array3t
from pyMathBitPrecise.bits3t import Bits3t
from pycocotb.basic_hdl_simulator.sim_utils import valueHasChanged


class BasicRtlSimProxy():
    """
    Signal proxy which manages the access to a memory in simulation

    :ivar ~.callbacks: list of sim processes which will be waken up if signal value is updated
    :ivar ~.sim: main simulator
    :ivar ~.name: name of property which is this proxy stored in on parent
    :ivar ~._name: signal name which was used in HDL
    :ivar ~._dtype: data type of this signal
    :ivar ~._origin: the object which was this proxy generated from
    :ivar ~._ag: agent which controlls this proxy
    :ivar ~.parent: parent object
    :ivar ~.def_val: value used for initialization of value (done on sim. startup)
    :ivar ~.val: actual value of signal
    :ivar ~.val_next: place for metainformations about next update
    """
    __slots__ = ["callbacks", "sim", "name", "_name", "parent",
                 "_dtype", "_origin", "_ag",
                 "def_val", "val", "val_next",
                 "simRisingSensProcs", "simFallingSensProcs", "simSensProcs"]
    BIT_t = Bits3t(1, False)

    def __init__(self, sim: "BasicRtlSimulator", parent, name, dtype, def_val):
        self.callbacks = []
        self.sim = sim
        self.parent = parent
        self.def_val = def_val
        self.val = dtype.from_py(None)
        self.val_next = None
        # properties used for simplified associations and debug in python
        self.name = name  # physical name
        self._name = name  # logical name
        self._dtype = dtype  # type notation for python
        self._origin = None  # signal object which this proxy substitutes
        self._ag = None  # simulation agent which drive or monitor this signal
        self.simRisingSensProcs = set()
        self.simFallingSensProcs = set()
        self.simSensProcs = set()

    def init_def_val(self, *args, **kwargs):
        self.def_val = self._dtype.from_py(*args, **kwargs)
        return self

    def read(self):
        assert self.sim.read_only_not_write_only
        return self.val.__copy__()

    def write(self, val):
        assert not self.sim.read_only_not_write_only
        t = getattr(val, "_dtype", None)
        if t is None:
            val = self._dtype.from_py(val)
        else:
            val = self._dtype.from_py(
                val.val,
                min(val.vld_mask,
                    self._dtype.all_mask())
            )
        if valueHasChanged(self.val, val):
            self.val = val
            self._propagate_changes(None)

    def wait(self, cb):
        self.callbacks.append(cb)
        self.sim.signals_checked_for_change.add(cb)

    def _apply_update(self, valUpdater):
        """
        Method called by simulator to update new value for this object
        """
        dirty_flag, new_val = valUpdater(self.val)
        assert new_val._dtype == self._dtype, (self, self.sim.time, new_val._dtype, self._dtype)
        if dirty_flag:
            self.val = new_val
            self._propagate_changes(valUpdater)

    def _propagate_changes(self, valUpdater):
        v = self.val
        sim = self.sim
        sim._updated_in_this_step.add(self)
        log = sim.logChange
        if log:
            log(sim.time, self, v, valUpdater)

        log = sim.logPropagation
        if log:
            log(sim, self, self.simSensProcs)

        # # run all sensitive processes
        for p in self.simSensProcs:
            sim._add_hdl_proc_to_run(self, p)

        # run write callbacks we have to create new list to allow
        # registering of new call backs in callbacks
        self.sim.pending_event_list.extend(self.callbacks)
        self.callbacks.clear()

        if self.simRisingSensProcs:
            if v.val or not v.vld_mask:
                if log:
                    log(sim, self, self.simRisingSensProcs)
                for p in self.simRisingSensProcs:
                    sim._add_hdl_proc_to_run(self, p)

        if self.simFallingSensProcs:
            if not v.val or not v.vld_mask:
                if log:
                    log(sim, self, self.simFallingSensProcs)
                for p in self.simFallingSensProcs:
                    sim._add_hdl_proc_to_run(self, p)

    def _onRisingEdge(self):
        v = self.val
        is_rising = self in self.sim._updated_in_this_step\
            and (v.val or not v.vld_mask)
        return self.BIT_t.from_py(int(is_rising), int(bool(v.vld_mask)))

    def _onFallingEdge(self):
        v = self.val
        is_falling = self in self.sim._updated_in_this_step\
            and (not v.val or not v.vld_mask)
        return self.BIT_t.from_py(int(is_falling), int(bool(v.vld_mask)))

    def __getitem__(self, index):
        if not isinstance(self._dtype, Array3t):
            raise TypeError("%r is not iterable because it uses type %r"
                            % (self, self._dtype))
        elif index < 0 or index >= self._dtype.size:
            raise IndexError(self._dtype.size, index)
        else:
            return BasicRtlSimProxyArrItem(self, index)

    def __len__(self):
        if not isinstance(self._dtype, Array3t):
            raise TypeError("%r is not iterable because it uses type %r"
                            % (self, self._dtype))
        else:
            return self._dtype.size

    def __repr__(self):
        return "<%s %r.%s %r %r>" % (self.__class__.__name__, self.parent,
                                     self.name, self.val, self.val_next)


class BasicRtlSimProxyArrItem():
    """
    Virtual proxy for an array item of BasicRtlSimProxy
    """

    def __init__(self, parent_proxy, item_index):
        self.parent_proxy = parent_proxy
        self.item_index = item_index
        self.sim = parent_proxy.sim
        self.parent = parent_proxy.parent
        self._dtype = parent_proxy._dtype.element_t

    def read(self):
        assert self.sim.read_only_not_write_only
        v = self.parent_proxy.val.val.get(self.item_index, None)
        if v is None:
            return self._dtype.from_py(None)

        return v.__copy__()



from os.path import dirname, abspath, join
from pycocotb.verilator.simulator_gen import verilatorCompile, \
    generatePythonModuleWrapper, loadPythonCExtensionFromFile

VERILOG_SRCS = dirname(abspath(__file__))


def format_accessible_signals(accessible_signals, top_name):
    _accessible_signals = []
    for signal_name, read_only, is_signed, type_width in  accessible_signals:
        # name as tuple, add top_name if is hierarchical
        if isinstance(signal_name, str):
            signal_phy_name = signal_name = (signal_name,)
        elif len(signal_name) > 1:
            signal_phy_name = (top_name, *signal_name)
        else:
            signal_phy_name = signal_name

        signal_phy_name = "__DOT__".join(signal_phy_name)

        read_only = int(read_only)
        is_signed = int(is_signed)
        if isinstance(type_width, int):
            type_width = (type_width,)

        _accessible_signals.append((signal_name, signal_phy_name, read_only, is_signed, type_width))       

    return _accessible_signals

    
def build_sim(verilog_files, accessible_signals, tc, build_dir, top_name):
    sim_verilog = [join(VERILOG_SRCS, f) for f in verilog_files]
    verilatorCompile(sim_verilog, build_dir)

    accessible_signals = format_accessible_signals(accessible_signals, top_name)
    module_file_name = generatePythonModuleWrapper(
        top_name, top_name,
        build_dir,
        accessible_signals)

    sim_module = loadPythonCExtensionFromFile(module_file_name, top_name)
    sim_cls = getattr(sim_module, top_name)

    simInstance = sim_cls()
    io = simInstance.io
    for sigName, sigPhyName, _, _, _ in accessible_signals:
        _io = io
        for n in sigName:
            _io = getattr(_io, n)
        tc.assertEqual(_io.name, sigName[-1])

    return simInstance



from os.path import join
from tempfile import TemporaryDirectory
import unittest

from pycocotb.constants import CLK_PERIOD
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.tests.common import build_sim
from pycocotb.triggers import Timer, WaitCombRead, WaitWriteOnly


class VerilatorWireTC(unittest.TestCase):
    """
    Simple test of verilator simulation wrapper functionality
    """

    def build_sim(self, build_dir, DW, top_name):
        accessible_signals = [
            # (signal_name, read_only, is_signed, type_width)
            ("inp", 0, 0, DW),
            ("outp", 1, 0, DW),
        ]
        files = [top_name + ".v", ]
        return build_sim(files, accessible_signals, self, build_dir, top_name)

    def _test_sim_wire(self, DW, test_data):
        # build_dir = "tmp"
        # if True:
        with TemporaryDirectory() as build_dir:
            rtl_sim = self.build_sim(build_dir, DW, "wire%d" % DW)
            io = rtl_sim.io
            sim = HdlSimulator(rtl_sim)

            r_data = []

            def data_collect():
                for d_ref in test_data:
                    yield WaitCombRead()
                    d = io.outp.read()
                    d = int(d)
                    r_data.append(d)
                    self.assertEqual(d, d_ref)
                    yield Timer(CLK_PERIOD)

            def data_feed():
                for d in test_data:
                    yield WaitWriteOnly()
                    io.inp.write(d)
                    yield Timer(CLK_PERIOD)

            rtl_sim.set_trace_file(join(build_dir, "wire%d.vcd" % DW), -1)
            sim.run(int(CLK_PERIOD * (len(test_data) + 0.5)),
                    extraProcesses=[
                        data_collect(),
                        data_feed()
                        ]
                    )
            self.assertEqual(len(r_data), len(test_data))

    def test_wire2(self):
        data = [1, 2, 3, 0, 2, 1, 2, 0, 2]
        self._test_sim_wire(2, data)

    def test_wire64(self):
        data = [1 << x for x in range(63)]
        self._test_sim_wire(64, data)

    def test_wire128(self):
        data = [1 << x for x in range(127)]
        self._test_sim_wire(128, data)


if __name__ == "__main__":
    suite = unittest.TestSuite()
    # suite.addTest(VerilatorWireTC('test_wire64'))
    suite.addTest(unittest.makeSuite(VerilatorWireTC))
    runner = unittest.TextTestRunner(verbosity=3)
    runner.run(suite)



"""
Agents for UART (COM, serial, Universal Asynchronous Receiver-Transmitter) interface

As this interface is asynchronous it does not need care clock signals or resets.
And BAUD has to be set correctly for correct functionality.
"""
from collections import deque
from typing import Tuple

from pycocotb.agents.base import AgentBase
from pycocotb.constants import Time
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.triggers import Timer, WaitTimeslotEnd


# constants for most common baud rates
BAUD_9600 = 9600
BAUD_19200 = 19200
BAUD_38400 = 38400
BAUD_57600 = 57600
BAUD_115200 = 115200
# constant which should be use when same baud is required but value is not important
BAUD_DEFAULT = BAUD_115200


class UartDataAgent(AgentBase):
    """
    :ivar ~.char_buff: the buffer used for bits while sending the character
    """
    START_BIT = 0
    STOP_BIT = 1

    def __init__(self, sim: HdlSimulator, intf: "RtlSignal", baud: int):
        super(UartDataAgent, self).__init__(sim, intf)
        self.set_baud(self, baud)
        self.char_buff = []
        self.data = deque()

    def set_baud(self, baud: int):
        self.bit_period = Time.s // baud
        assert self.bit_period > 2, baud

    def recieve_text(self) -> str:
        """
        :return: data of this agent (collected form interface) as str
        """

    def send_text(self, text: str):
        """
        Add text to agent data (agent will send it as soon as ready)
        """

    def monitor(self):
        half_period = Timer(self.bit_period // 2)
        period = Timer(self.bit_period)
        intf = self.intf

        yield half_period

        while True:
            while True:
                if not self.getEnable():
                    yield period
                    continue

                yield WaitTimeslotEnd()
                d = intf.read()
                if int(d) == self.START_BIT:
                    break
                else:
                    yield half_period

            for _ in range(8):
                yield period
                yield WaitTimeslotEnd()
                d = intf.read()
                self.char_buff.append(d)

            d = int(d)
            if d == self.STOP_BIT:
                # correctly received char
                self.char_buff.clear()
                ch = 0
                for b in reversed(self.char_buff):
                    ch <<= 1
                    ch |= int(b)
                yield period
            else:
                # received data is in wrong format, discard it
                self.char_buff.clear()
                yield half_period

    def driver(self):
        half_period = Timer(self.bit_period // 2)
        period = Timer(self.bit_period)
        intf = self.intf
        yield half_period
        while True:
            if self.getEnable() and self.data:
                ch = self.data.popleft()
                intf.write(self.START_BIT)
                yield period
                for i in range(8):
                    intf.write((ch >> i) & 0b1)
                    yield period
                intf.write(self.STOP_BIT)
            yield period


class UartRxTxAgent(AgentBase):

    def __init__(self, intf: Tuple["RtlSignal", "RtlSignal"], baud: int):
        super(UartRxTxAgent, self).__init__(intf)
        rx, tx = intf
        self.rx = UartDataAgent(rx, baud)
        self.tx = UartDataAgent(tx, baud)

    def set_baud(self, baud: int):
        self.rx.set_baud(baud)
        self.tx.set_baud(baud)

    def getDrivers(self):
        return (
            *self.rx.getMonitors(),
            *self.tx.getDrivers(),
        )

    def getMonitors(self):
        return (
            *self.rx.getDrivers(),
            *self.tx.getMonitors(),
        )



from collections import deque
from typing import Tuple, Union, Optional

from pycocotb.agents.base import AgentWitReset, NOP, RX, TX
from pycocotb.agents.peripheral.tristate import TristateAgent, TristateClkAgent,\
    TristateSignal
from pycocotb.triggers import WaitCombStable, WaitWriteOnly, WaitCombRead,\
    WaitTimeslotEnd
from enum import Enum
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.process_utils import OnRisingCallbackLoop, OnFallingCallbackLoop


class I2C_MODE(Enum):
    STANDARD = 100000
    FAST = 400000
    HIGH_SPEED = 3400000


class I2C_ADDR(Enum):
    ADDR_7b = 7
    ADDR_10b = 10


def getBit(val, bitNo):
    """
    get specific bit from integer (little-endian)
    """
    return (val >> bitNo) & 1


class I2cAgent(AgentWitReset):
    """
    A base simulation agent for I2C interfaces. Note that as all devices are using tri-state interfaces.
    The agent has to know the data format which of device in order to interpret data correctly.
    https://www.nxp.com/docs/en/user-guide/UM10204.pdf
    https://dlbeer.co.nz/articles/i2c.html

    :note: The agent will support slave mode after it's address is set.

    Master agent transaction formats:
          [((READ, ACK/NACK) | (WRITE, VALUE, ACK/NACK) | (REPEATED_START, duration in I2C clk))*]
          (ACK/NACK values are there to check the slave response)
    Slave agent uses the onWrite/onRead callbacks

    :attention: multi-master arbitration not implemented
    :note: The driver/monitor difference is related to directions of signals but as the used interface
        is tri-state interface bouth agents can be SLAVE/master
    :note: typical I2C command: START, A6-0, R/W, ACK,  ...  D7-0, ACK, STOP
    :note: each agent is master, but none is active if it does not have transactions to do so
    :note: slave has to check address in function because some devices may actually use
        part of address as opt code etc. because of this this agent needs to be more generic

    :ivar ~.data_m: the buffer for transaction for master
    :ivar ~.data_m_read: the buffer for read data by master
    :ivar ~.start: flag, if True the master should send I2C START
    :ivar ~.stop: flag, if True the master should send I2C STOP

    :ivar ~.bit_cntrl: the buffer for commands for specific SDA value
                    (intermediate instructions)
    :ivar ~.bit_cntrl_rx: the buffer for a bit values of SDA
                    (intermediate read value)
    """

    ADDR_PREFIX_10b = 0b11110000

    ACK = 0  # send by reciever
    NACK = 1

    READ = 1
    WRITE = 0

    START = "START"
    RESTART = START
    STOP = "STOP"

    def __init__(self, sim: HdlSimulator, intf: Tuple[TristateSignal, TristateSignal],
                 rst: Tuple["RtlSignal", bool]):
        """
        :param: intf i2c interface, tuple (scl, sda) = (clock, data),
            sda/sdc are tri-state interfaces represented by i, o, t signals
        """
        AgentWitReset.__init__(self, sim, intf, rst)
        self.data_m = deque()
        self.data_m_read = []
        self.bit_cntrl = deque()  # type: Deque[ Tuple[Union[RX, TX], Optional[int]] ]
        self.bit_cntrl_rx = deque()  # type: Deque[Union[START, STOP, int]]
        self.start = True
        self.sda = TristateAgent(sim, intf[1], rst)
        self.sda.collectData = False
        self.sda.selfSynchronization = False
        self.slave = False
        self.bit_index = None

    def hasTransactionPending(self):
        return (
            self.data_m
            or self.data_m_read
            or self.bit_cntrl
            or self.bit_cntrl_rx)

    def _transmit_byte(self, val: int, ack_for_check: Optional[bool]):
        for i in range(8):
            b = getBit(val, 7-i)
            self.bit_cntrl.append((TX, b))
        self.bit_cntrl.append((RX, ack_for_check))
        yield from self._wait_until_command_completion()
        assert len(self.bit_cntrl_rx) == 1
        return self.bit_cntrl_rx.pop()

    def _receive_byte(self, ack: bool):
        """
        :note: If master is a reciever the ack means that it wishes to recieve next byte
        """
        for _ in range(8):
            self.bit_cntrl.append((RX, None))
        self.bit_cntrl.append((TX, int(ack)))
        yield from self._wait_until_command_completion()
        assert len(self.bit_cntrl_rx) == 8
        b = 0
        for _b in self.bit_cntrl_rx:
            b <<= 1
            b |= _b
        return b

    def _wait_until_command_completion(self):
        # wait untill command get processed
        while self.bit_cntrl:
            yield WaitCombRead()
            if self.bit_cntrl:
                yield WaitTimeslotEnd()

    def execute_master_transaction(self):
        trans = self.data.pop()
        for t in trans:
            m = t[0]
            if m == RX:
                b = yield from self._receive_byte(t[1])
                raise NotImplementedError(b)
            elif m == TX:
                ack = yield from self._transmit_byte(b[1], b[2])
                raise NotImplementedError(ack)
            elif m == self.START:
                raise NotImplementedError()
            elif m == self.STOP:
                raise NotImplementedError()
            else:
                ValueError(m)
        self.stop = True

    def monitor_on_read(self, addr):
        raise NotImplementedError()
    def monitor_on_write(self, addr):
        raise NotImplementedError()

    def execute_slave_transaction(self):
        if self.ADDR_BITS == I2C_ADDR.ADDR_7b:
            yield from self._receive_byte(1)
            addr = self.data_rx.pop()
            rw = addr & 0b1
            addr >>= 1
        elif self.ADDR_BITS == I2C_ADDR.ADDR_10b:
            yield from self._receive_byte(1)
            addrLow = self.data_rx.pop()
            addrLow = int(addrLow)
            rw = addrLow & 0b1
            addrLow >>= 1
            assert addr & 0b11111000 == self.ADDR_PREFIX_10b
            yield from self._receive_byte(1)
            addrHigh = self.data_rx.pop()
            addrHigh = int(addrHigh)
            addr = (addr << 8) | addrHigh

        if rw == self.READ:
            yield from self.monitor_on_read(addr)
        else:
            yield from self.monitor_on_write(addr)

    def startListener(self):
        # SDA->0 and SCL=1
        if self.start:
            self.bit_index = 0
            self.bit_cntrl_rx.append(self.START)
            self.start = False
            # yield sexecute_master_transactionelf.execute_master_transaction()

        return
        yield

    def startSender(self):
        # SDA->0 and SCL=1
        if self.start:
            self.bit_index = 0
            self.sda._write(0)
            self.start = False

        return
        yield

    def getMonitors(self):
        sim = self.sim
        scl = self.intf[0]
        self.scl = TristateClkAgent(
            sim, scl, (self.rst, self.rstOffIn),
        )
        self.monitor = OnRisingCallbackLoop(
            sim, scl.i, self.monitor, self.getEnable)
        self.startListener = OnFallingCallbackLoop(
            sim, scl.i, self.startListener, self.getEnable)

        return (
            self.monitor(),
            self.startListener(),
            *self.sda.getMonitors(),
            *self.scl.getMonitors()
        )

    def getDrivers(self):
        sim = self.sim
        scl = self.intf[0]
        driver = self.driver
        self.scl = TristateClkAgent(
            sim, scl, (self.rst, self.rstOffIn),
        )
        self.driver = OnRisingCallbackLoop(
            sim, scl.i, self.driver, self.getEnable)
        self.startSender = OnFallingCallbackLoop(
            sim, scl.i, self.startSender, self.getEnable)
        self.scl.setEnable(False, None)
        return (
            driver(),  # initialization of the interface
            self.driver(),
            self.startSender(),
            *self.sda.getDrivers(),
            *self.scl.getDrivers()
        )

    def monitor(self):
        # now intf.sdc is rising
        yield WaitCombRead()
        # wait on all agents to update values and on
        # simulator to apply them
        if self.sim.now > 0 and self.notReset():
            if self.bit_index != 8:
                self.bit_index += 1
                yield WaitCombStable()
                v = self.sda.i.read()
                self.bit_cntrl_rx.append(v)
            else:
                yield WaitWriteOnly()
                self.sda._write(self.ACK)

    def driver(self):
        # now intf.sdc is rising
        # prepare data for next clk
        yield WaitWriteOnly()
        if self.bits:
            b = self.bits.popleft()
            if b == self.START:
                self.start = True
                return
            elif b == self.STOP:
                self.bit_index = None
                self.stop = True
                return
        else:
            b = NOP

        self.sda._write(b)

    def setEnable(self, en):
        """
        """
        # If there is no pending transaction no pause is required
        if not self.start and not self.stop and not self.bit_cntrl:
            super(I2cAgent, self).setEnable(en)
        else:
            # wait until clock
            raise NotImplementedError()



from tempfile import TemporaryDirectory
import unittest

from pycocotb.agents.clk import ClockAgent
from pycocotb.agents.handshaked import HandshakedAgent
from pycocotb.agents.rst import PullUpAgent
from pycocotb.constants import CLK_PERIOD
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.tests.common import build_sim
from pycocotb.tests.verilatorHierarchy_test import VerilatorHierarchyTC


class HSAg(HandshakedAgent):

    def set_valid(self, val: bool):
        return self.intf[0].write(val)

    def get_valid(self):
        return self.intf[0].read()

    def get_ready(self):
        return self.intf[1].read()

    def set_ready(self, val: bool):
        self.intf[1].write(val)

    def set_data(self, data):
        self.intf[2].write(data)

    def get_data(self):
        return self.intf[2].read()


def generate_handshaked_agents(sim, io, clk, rst_n):

    din_ag = HSAg(sim, (io.dataIn_vld, io.dataIn_rd, io.dataIn_data),
                  clk, (rst_n, True))
    dout_ag = HSAg(sim, (io.dataOut_vld, io.dataOut_rd, io.dataOut_data),
                   clk, (rst_n, True))

    return din_ag, dout_ag


class VerilatorHandshakedWireTC(unittest.TestCase):
    """
    Tests of Handshaked agent
    """

    def hw_build(self, build_dir):
        """
        Build simulator for HandshakedWire.v in specified dir
        """
        accessible_signals = [
            # (signal_name, read_only, is_signed, type_width)
            ("clk", 0, 0, 1),
            ("rst_n", 0, 0, 1),
            ("dataIn_data", 0, 0, 8),
            ("dataIn_rd", 0, 0, 1),
            ("dataIn_vld", 1, 0, 1),

            ("dataOut_data", 1, 0, 8),
            ("dataOut_rd", 0, 0, 1),
            ("dataOut_vld", 1, 0, 1),
        ]
        verilog_files = ["HandshakedWire.v"]
        return build_sim(verilog_files, accessible_signals, self,
                         build_dir, "HandshakedWire")

    def _test_pass_data(self, initFn):
        # build_dir = "tmp"
        # if True:
        with TemporaryDirectory() as build_dir:
            rtl_sim = self.hw_build(build_dir)
            io = rtl_sim.io
            sim = HdlSimulator(rtl_sim)

            din_ag, dout_ag = generate_handshaked_agents(
                sim, io, io.clk, io.rst_n)

            extra_procs = initFn(sim, din_ag, dout_ag)
            if extra_procs is None:
                extra_procs = []

            proc = [
                *ClockAgent(sim, io.clk, CLK_PERIOD).getDrivers(),
                *PullUpAgent(sim, io.rst_n, CLK_PERIOD).getDrivers(),
                *din_ag.getDrivers(),
                *dout_ag.getMonitors(),
                *extra_procs,
            ]
            sim.run(int(CLK_PERIOD * 20.5), extraProcesses=proc)
            return sim, din_ag, dout_ag

    def test_nop(self):

        def init(sim, din_ag, dout_ag):
            pass

        _, din_ag, dout_ag = self._test_pass_data(init)
        self.assertSequenceEqual(din_ag.data, [])
        self.assertSequenceEqual(dout_ag.data, [])

    def test_simple_data(self):
        ref_data = [1, 2, 3, 4, 5]

        def init(sim, din_ag, dout_ag):
            din_ag.data.extend(ref_data)

        _, din_ag, dout_ag = self._test_pass_data(init)

        self.assertSequenceEqual(din_ag.data, [])
        self.assertSequenceEqual(dout_ag.data, ref_data)

    def test_fifo(self):
        ref_data = [1, 2, 3, 4, 5]

        def init(sim, din_ag, dout_ag):
            # sim.rtl_simulator.set_trace_file("tmp/handshaked_fifo.vcd", -1)
            din_ag.data.extend(ref_data)

        hw_build = self.hw_build

        def hf_build(build_dir):
            return VerilatorHierarchyTC.build_handshaked_fifo(self, build_dir)

        try:
            self.hw_build = hf_build
            _, din_ag, dout_ag = self._test_pass_data(init)
            self.assertSequenceEqual(din_ag.data, [])
            self.assertSequenceEqual(dout_ag.data, ref_data)
        finally:
            self.hw_build = hw_build


if __name__ == "__main__":
    suite = unittest.TestSuite()
    # suite.addTest(VerilatorHandshakedWireTC('test_fifo'))
    suite.addTest(unittest.makeSuite(VerilatorHandshakedWireTC))
    runner = unittest.TextTestRunner(verbosity=3)
    runner.run(suite)



from pycocotb.agents.base import AgentBase
from pycocotb.constants import CLK_PERIOD
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.process_utils import CallbackLoop
from pycocotb.triggers import Timer, WaitWriteOnly, WaitCombRead


class ClockAgent(AgentBase):
    """
    Simulation agent for :class:`hwt.interfaces.std.Clk` interface

    * In driver mode oscillates at frequency specified by period

    * In monitor driver captures tuples (time, nextVal) for each change
        on signal (nextVal is 1/0/None)

    :ivar ~.period: period of signal to generate
    :ivar ~.initWait: time to wait before starting oscillation
    """

    def __init__(self, sim: HdlSimulator, intf: "RtlSignal", period: int=CLK_PERIOD):
        super(ClockAgent, self).__init__(sim, intf)
        assert isinstance(period, int)
        self.period = period
        self.initWait = 0
        self.monitor = CallbackLoop(sim, self.intf, self.monitor, self.getEnable)

    def driver(self):
        assert isinstance(self.period, int)
        assert isinstance(self.initWait, int)
        sig = self.intf
        yield WaitWriteOnly()
        sig.write(0)
        yield Timer(self.initWait)

        while True:
            halfPeriod = self.period // 2

            yield Timer(halfPeriod)
            yield WaitWriteOnly()
            sig.write(1)

            yield Timer(halfPeriod)
            yield WaitWriteOnly()
            sig.write(0)

    def getMonitors(self):
        self.last = (-1, None)
        self.data = []
        return super(ClockAgent, self).getMonitors()

    def monitor(self):
        assert isinstance(self.period, int)
        assert isinstance(self.initWait, int)
        yield WaitCombRead()
        v = self.intf.read()
        try:
            v = int(v)
        except ValueError:
            v = None

        now = self.sim.now
        last = self.last

        _next = (now, v)
        if last[0] == now:
            if last[1] is not v:
                # update last value
                last = (now, v)
                self.last = last
                if self.data:
                    self.data[-1] = last
                else:
                    self.data.append(last)
            else:
                return
        else:
            self.data.append(_next)

        self.last = _next



from os.path import join
from tempfile import TemporaryDirectory
import unittest

from pycocotb.agents.clk import ClockAgent
from pycocotb.agents.rst import PullDownAgent, PullUpAgent
from pycocotb.constants import CLK_PERIOD
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.tests.common import build_sim
from pycocotb.tests.example_agents import get_clk_driver, get_rst_driver, \
    get_pull_up_driver, get_sync_sig_monitor, get_pull_up_driver_with_reset, \
    get_sync_pull_up_driver_with_reset
from pycocotb.triggers import Timer, WaitCombStable


REF_DATA = [
    (15000, 0),
    (25000, 1),
    (35000, 2),
    (45000, 3),
    (55000, 0),
    (65000, 1),
    (75000, 2),
    (85000, 3),
    (95000, 0)
]


class VerilatorCntrTC(unittest.TestCase):
    """
    Simple test of verilator simulation wrapper functionality
    """

    def cntr_build(self, build_dir):
        """
        Build simulator for Cntr.v in specified dir
        """
        accessible_signals = [
            # (signal_name, read_only, is_signed, type_width)
            ("clk", 0, 0, 1),
            ("en", 0, 0, 1),
            ("rst", 0, 0, 1),
            ("val", 1, 0, 2),
        ]
        verilog_files = ["Cntr.v"]
        return build_sim(verilog_files, accessible_signals, self, build_dir, "Cntr")

    def test_dual_build(self):
        """
        Test if simulation doe not interfere between each other
        """
        with TemporaryDirectory() as build_dir0, TemporaryDirectory() as build_dir1:
            sim0 = self.cntr_build(build_dir0)
            sim1 = self.cntr_build(build_dir1)
            self.assertIsNot(sim0, sim1)

    def test_sim_cntr(self):
        """
        Time synchronized monitors (val) and drivers (clk, rst, en)
        """
        # build_dir = "tmp"
        # if True:
        with TemporaryDirectory() as build_dir:
            rtl_sim = self.cntr_build(build_dir)
            io = rtl_sim.io
            data = []

            def data_collector():
                yield Timer(CLK_PERIOD // 2)
                assert sim.now == CLK_PERIOD // 2

                val = io.val
                while True:
                    yield Timer(CLK_PERIOD)
                    yield WaitCombStable()
                    data.append((sim.now, int(val.read())))

            sim = HdlSimulator(rtl_sim)
            # rtl_sim.set_trace_file("cntr.vcd", -1)
            procs = [
                get_clk_driver(sim, io.clk, CLK_PERIOD),
                get_rst_driver(sim, io.rst, CLK_PERIOD),
                get_pull_up_driver(sim, io.en, CLK_PERIOD),
                data_collector()
            ]
            sim.run(int(CLK_PERIOD * 10.5), extraProcesses=procs)

            self.assertSequenceEqual(data, REF_DATA)

    def test_sim_cntr2(self):
        """
        Clock dependency on clk
            * monitor of val
        Simulation step restart due write after reset read
        """
        # build_dir = "tmp"
        # if True:
        with TemporaryDirectory() as build_dir:
            rtl_sim = self.cntr_build(build_dir)
            io = rtl_sim.io
            sim = HdlSimulator(rtl_sim)
            data = []

            # rtl_sim.set_trace_file("cntr.vcd", -1)
            procs = [
                get_clk_driver(sim, io.clk, CLK_PERIOD),
                get_rst_driver(sim, io.rst, CLK_PERIOD),
                get_pull_up_driver(sim, io.en, CLK_PERIOD),
                get_sync_sig_monitor(sim, io.val, io.clk, io.rst, data)
            ]
            sim.run(int(CLK_PERIOD * 10.5), extraProcesses=procs)

            self.assertSequenceEqual(data, REF_DATA)

    def test_sim_cntr_pull_up_reset(self):
        """
        Clock dependency on clk
            * monitor of val
        Simulation step restart due write after reset read
        """
        # build_dir = "tmp"
        # if True:
        with TemporaryDirectory() as build_dir:
            rtl_sim = self.cntr_build(build_dir)
            io = rtl_sim.io
            data = []

            sim = HdlSimulator(rtl_sim)
            # rtl_sim.set_trace_file("cntr.vcd", -1)
            proc = [
                get_clk_driver(sim, io.clk, CLK_PERIOD),
                get_rst_driver(sim, io.rst, CLK_PERIOD),
                get_pull_up_driver_with_reset(sim, io.en, io.rst, CLK_PERIOD),
                get_sync_sig_monitor(sim, io.val, io.clk, io.rst, data)
            ]
            sim.run(int(CLK_PERIOD * 10.5), extraProcesses=proc)

            self.assertSequenceEqual(data, REF_DATA)

    def test_sim_cntr_sync_pull_up_reset(self):
        """
        Clock dependency on clk
            * driver of en
            * monitor of val
        Simulation step restart due write after reset read
        """
        # build_dir = "tmp"
        # if True:
        with TemporaryDirectory() as build_dir:
            rtl_sim = self.cntr_build(build_dir)
            io = rtl_sim.io
            data = []

            sim = HdlSimulator(rtl_sim)
            # rtl_sim.set_trace_file(join(build_dir, "cntr.vcd"), -1)
            proc = [
                get_clk_driver(sim, io.clk, CLK_PERIOD),
                get_rst_driver(sim, io.rst, CLK_PERIOD),
                get_sync_pull_up_driver_with_reset(sim, io.en, io.clk, io.rst),
                get_sync_sig_monitor(sim, io.val, io.clk, io.rst, data)
            ]
            sim.run(int(CLK_PERIOD * 10.5), extraProcesses=proc)

            self.assertSequenceEqual(data, REF_DATA)

    def test_sim_normal_agents(self):
        # build_dir = "tmp"
        # if True:
        with TemporaryDirectory() as build_dir:
            rtl_sim = self.cntr_build(build_dir)
            io = rtl_sim.io
            sim = HdlSimulator(rtl_sim)
            data = []
            procs = [
                *ClockAgent(sim, io.clk).getDrivers(),
                *PullDownAgent(sim, io.rst).getDrivers(),
                *PullUpAgent(sim, io.en, initDelay=CLK_PERIOD).getDrivers(),
                get_sync_sig_monitor(sim, io.val, io.clk, io.rst, data)
            ]
            rtl_sim.set_trace_file(join(build_dir, "cntr.vcd"), -1)
            sim.run(int(CLK_PERIOD * 10.5), extraProcesses=procs)

            self.assertSequenceEqual(data, REF_DATA)


if __name__ == "__main__":
    suite = unittest.TestSuite()
    # suite.addTest(VerilatorCntrTC('test_sim_cntr_sync_pull_up_reset'))
    suite.addTest(unittest.makeSuite(VerilatorCntrTC))
    runner = unittest.TextTestRunner(verbosity=3)
    runner.run(suite)





class Time():
    """
    Time units
    """
    ps = 1
    ns = 1000
    us = ns * 1000
    ms = us * 1000
    s = ms * 1000


# default clk period
# value is not important. But is is nice to have same value over all designs.
CLK_PERIOD = 10 * Time.ns



"""
Pitfalls of delta-step based HDL simulators

* HDL simulator
    * manages communication between the RTL simulator and the simulation environment
* RTL simulator
    * simulates the circuit itself

* Simulation environment (=simulator).
    * HDL simulator + RTL simulator + user specified sim. processes
* User specified processes
    * Can read all signals in RTL simulator
    * Can write and wait for event only on top level signals.
    * Can preinitialize values of signals. (e.g. initialize memories before simulation start)

* Problem is how to ensure correct order of operations in RTL simulator
  in user specified processes which are controlling RTL simulator
  while keeping simulation simple and fast.

    * Requirements of UVM like interface agents.

        * Agents needs to be able to:

            * Perform combinational loop (e.g. tristate pull-up in I2C)

            * Wait for (before and after) edge of clock signal (e.g. BRAM port)

                * Clock signal can be driven directly by sim.
                  If this is a case synchronization is simple because simulator
                  and user knows when the event happens (before call delta step).

                * If clock signal is generated by circuit it is problematic because
                  value of clock signal is updated in simulation step together
                  with another signal updates. This can result in incorrect
                  event order and special care is required.

            * Perform combinational loop for clock signals.

                * Value has to be updated in same step.

                    * This practically means that the event can appear on multiple signals
                      at once.

        * Agent is full description of interface protocol and should be universal
          and extensible.

    * Combinational loops between simulator and RTL simulator.

        * If only one signal is updated per delta step there is no problem.

        * However if multiple signals signals (especially clock signals)
          are updated in same delta step combinational loop does not
          have to be resolved before some event dependent (e.g. FF) update
          This results in invalid resolution.

            * Updates of combinational logic has to be resolved before update
              of memory elements and dependent logic.

    * Clock generated by RTL simulator.

        * This is problem for non-pure DES simulators like Verilator.

        * Output signals from RTL simulator does not have to have
          values which they have when clock event was triggered, solution:


            * Manually add registers in to circuit to hold correct value.

            * Run simulation process which waits on event on specified signal
              exactly in time of the event. Problem is that in this state some
              values can be undefined and read only access is required.
"""

from inspect import isgenerator
from typing import List

from pycocotb.simCalendar import SimTimeSlot, SimCalendar, DONE
from pycocotb.triggers import Event, raise_StopSimulation, \
    StopSimumulation, Action


# similar to https://github.com/potentialventures/cocotb/blob/master/cocotb/scheduler.py
class HdlSimulator():
    """
    This simulator simulates the communication between circuit simulator
    and simulation processes which are driving the circuit simulation.
    Simulation processes are usually provided by simulation agents or user.

    :ivar ~.now: actual simulation time
    :ivar ~._events: heap of simulation events and processes
    :ivar ~.rtl_simulator: circuit simulator used for simulation of circuit itself
    """

    def __init__(self, rtl_simulator):
        self.rtl_simulator = rtl_simulator
        self.now = 0
        self._events = SimCalendar()
        self._current_time_slot = None  # type: SimTimeSlot
        self._current_event_list = None  # type: List

        schedule = self._events.push

        # def schedule(*args):
        #     assert self.now <= args[0]
        #     print(self.now, "sched:", *args)
        #     self._events.push(*args)
        #
        self.schedule = schedule

    def _run_process(self, process):
        """
        Execute process and process it's requests
        """
        # run process or activate processes dependent on Event
        for ev in process:
            # if process requires waiting put it back in queue
            if isinstance(ev, Action):
                if not ev.applyProcess(self, process):
                    break
            elif isinstance(ev, Event):
                ev.applyProcess(self, process)
                break
            elif isgenerator(ev):
                # else this process spotted new process
                # and it has to be put in queue
                self._schedule_proc_now(ev)
            else:
                raise ValueError(ev)

    def _eval_rtl_events(self):
        """
        Run processes/events triggered by RTL simulator
        """
        rtl_sim = self.rtl_simulator
        rtl_pending_event_list = rtl_sim.pending_event_list
        if rtl_pending_event_list:
            # proper solution is to put triggered events to sim.
            # calendar with urgent priority  but we evaluate
            # it directly because of performance
            for _process in rtl_pending_event_list:
                if not isgenerator(_process):
                    _process = _process(self)

                self._run_process(_process)
            rtl_pending_event_list.clear()

    def _run_event_list(self, events):
        """
        Run block of events or processes
        """
        if events is not None:
            self._current_event_list = events
            for ev in events:
                # process is Python generator or Event
                if isinstance(ev, Event):
                    for p in ev:
                        self._run_process(p)
                else:
                    self._run_process(ev)

        self._current_event_list = None

    def run(self, until: int, extraProcesses=[]) -> None:
        """
        Run simulation for a specified time

        :note: Can be used to run simulation again after it ends from time when it ends.
        :note: Simulator restart is performed by new instantiation of the simulator.
        """

        assert until >= self.now, (until, self.now)
        if until == self.now:
            return

        now = self.now
        time_slot = SimTimeSlot()
        time_slot.write_only = []
        for proc in extraProcesses:
            assert isgenerator(proc), proc
            time_slot.write_only.append(proc)
        # add handle to stop simulation
        self.schedule(now, time_slot)

        end_time_slot = SimTimeSlot()
        end_time_slot.write_only = [raise_StopSimulation(self), ]
        self.schedule(now + until, end_time_slot)

        next_time_slot = self._events.pop
        rtl_sim = self.rtl_simulator
        _run_event_list = self._run_event_list
        END = rtl_sim.END_OF_STEP
        try:
            # for all events
            while True:
                now, time_slot = next_time_slot()
                self._current_time_slot = time_slot
                assert now >= self.now, (now, time_slot)
                rtl_sim.time = self.now = now

                # run preinitialization of sim. environment
                _run_event_list(time_slot.timeslot_begin)
                time_slot.timeslot_begin = DONE

                # run resolution of combinational lopps
                first_run = True
                while first_run or time_slot.write_only:
                    _run_event_list(time_slot.write_only)
                    time_slot.write_only = None
                    s = rtl_sim.eval()

                    assert s == rtl_sim.COMB_UPDATE_DONE, (self.now, s)
                    if time_slot.comb_read is None:
                        self._current_event_list = time_slot.comb_read = []
                    else:
                        self._current_event_list = time_slot.comb_read
                    self._eval_rtl_events()

                    _run_event_list(time_slot.comb_read)
                    time_slot.comb_read = None

                    if time_slot.write_only is not None:
                        # we have to reevaluate the combinational logic
                        # if write in this time stamp is required
                        rtl_sim.reset_eval()
                    first_run = False

                time_slot.write_only = DONE
                time_slot.comb_read = DONE

                # run evaluation of rest of the circuit
                while not rtl_sim.read_only_not_write_only:
                    rtl_sim.eval()
                    if rtl_sim.pending_event_list:
                        if time_slot.comb_read is None:
                            self._current_event_list = time_slot.comb_stable = []
                        else:
                            self._current_event_list = time_slot.comb_stable
                        self._eval_rtl_events()

                _run_event_list(time_slot.comb_stable)
                time_slot.comb_stable = DONE

                while True:
                    ret = rtl_sim.eval()
                    if rtl_sim.pending_event_list:
                        if time_slot.mem_stable is None:
                            self._current_event_list = time_slot.mem_stable = []
                        else:
                            self._current_event_list = time_slot.mem_stable
                        self._eval_rtl_events()
                    if ret == END:
                        break
                _run_event_list(time_slot.mem_stable)
                time_slot.mem_stable = DONE

                _run_event_list(time_slot.timeslot_end)
                time_slot.timeslot_end = DONE
                rtl_sim.set_write_only()

        except StopSimumulation:
            pass
        finally:
            rtl_sim.finalize()
        # to allow tesbenches to peek in to DUT after sim ended
        rtl_sim.read_only_not_write_only = True

    def _schedule_proc_now(self, ev):
        assert isinstance(ev, (Action, Event)) or isgenerator(ev), ev
        self._current_event_list.append(ev)

    def _schedule_proc(self, time: int, ev) -> None:
        assert isinstance(ev, (Action, Event)) or isgenerator(ev), ev
        if self.now == time:
            self._current_event_list.append(ev)
        else:
            ts = self._events.get(time, None)
            if ts is None:
                ts = SimTimeSlot()
                self.schedule(time, ts)
            if ts.write_only is None:
                ts.write_only = []
            ts.write_only.append(ev)



from pycocotb.simCalendar import DONE


class Event():
    """
    Simulation event

    Container of processes to wake

    :param process_to_wake: list of sim. processes (generator instances)
        to wake when this event is triggered
    """
    __slots__ = ["debug_name", "process_to_wake"]

    def __init__(self, debug_name=None):
        self.debug_name = debug_name
        self.process_to_wake = []

    def __iter__(self):
        return iter(self.process_to_wake)

    def destroy(self):
        self.process_to_wake = None

    def applyProcess(self, sim, process):
        self.process_to_wake.append(process)

    def __repr__(self):
        if self.debug_name is None:
            return super(Event, self).__repr__()
        else:
            return "<Event {} {:#018x}>".format(self.debug_name, id(self))


class StopSimumulation(BaseException):
    """
    Exception raised from handle in simulation to stop simulation
    """
    pass


def raise_StopSimulation(sim):
    """
    Simulation process used to stop simulation
    """
    raise StopSimumulation()
    return
    yield


class Action():

    def applyProcess(self, sim, process):
        raise NotImplementedError()


# [TODO] RisingEdge/FallingEdge with support in c++ code
class Edge(Action):
    """
    :note: if multiple signals specified the process will be triggered on first
        edge on any signal, once at most
    """

    def __init__(self, *signals: "RtlSignal"):
        self.signals = signals

    def applyProcess(self, sim, process):
        if len(self.signals) > 1:

            def wrap():
                """
                Because we need to wake process only once
                and after ignore wake potentially caused by edges on other signals
                """
                yield process

            p = wrap()
        else:
            p = process

        for s in self.signals:
            s.wait(p)

        return False


class Timer(Action):
    """
    Container for wait time of processes

    next activation of process will be now + time
    """

    def __init__(self, time: int):
        assert isinstance(time, int)
        self.time = time

    def applyProcess(self, sim, process):
        sim._schedule_proc(sim.now + self.time, process)
        # return false to notify that the evaluation should be paused
        return False

    def __repr__(self):
        return "<Timer %r>" % (self.time)


class WaitWriteOnly(Action):

    def applyProcess(self, sim, process):
        t = sim._current_time_slot
        ev_list = t.write_only
        if ev_list is None:
            ev_list = t.write_only = []
        elif ev_list is DONE:
            raise AssertionError("Can not write in this time slot any more",
                                 sim.now, process, sim._current_time_slot)
        ev_list.append(process)
        return False


class WaitCombRead(Action):

    def applyProcess(self, sim, process):
        t = sim._current_time_slot
        ev_list = t.comb_read
        if ev_list is None:
            ev_list = t.comb_read = []
        elif ev_list is DONE:
            # use later read only event as the condition is satisfied
            ev_list = sim._current_event_list
        ev_list.append(process)
        return False


class WaitCombStable(Action):

    def applyProcess(self, sim, process):
        t = sim._current_time_slot
        ev_list = t.comb_stable
        if ev_list is None:
            ev_list = t.comb_stable = []
        elif ev_list is DONE:
            # use later read only event as the condition is satisfied
            ev_list = sim._current_event_list

        ev_list.append(process)
        return False


class WaitTimeslotEnd(Action):

    def applyProcess(self, sim, process):
        t = sim._current_time_slot
        ev_list = t.timeslot_end
        if ev_list is None:
            ev_list = t.timeslot_end = []
        elif ev_list == DONE:
            ev_list = sim._current_event_list
        ev_list.append(process)
        return False



import os
from setuptools import setup, find_packages
from setuptools.extension import Library
import sys
from os import path

ext_modules = []

if "--verilator" in sys.argv:
    sys.argv.remove("--verilator")

    COCOPY_SRC_DIR = os.path.join(
        os.path.dirname(__file__),
        "pycocotb", "verilator", "c_files")
    COCOPY_SRCS = [os.path.join(COCOPY_SRC_DIR, p)
                   for p in [
                             "signal_mem_proxy.cpp",
                             "signal_array_mem_proxy.cpp",
                             "sim_io.cpp",
                             "pycocotb_sim.cpp"]
                   ]
    #VERILATOR_ROOT = "/usr/local/share/verilator"
    VERILATOR_ROOT = "./verilator"

    VERILATOR_INCLUDE_DIR = os.path.join(VERILATOR_ROOT, "include")
    VERILATOR_SOURCES = [
        os.path.join(VERILATOR_INCLUDE_DIR, x)
        for x in ["verilated.cpp", "verilated_save.cpp", "verilated_vcd_c.cpp"]
    ]

    verilator_common = Library(
        "pycocotb.verilator.common",
        sources=COCOPY_SRCS + VERILATOR_SOURCES,
        extra_compile_args=["-std=c++11", "-I" + VERILATOR_INCLUDE_DIR],
    )
    ext_modules.append(verilator_common)


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pycocotb',
    version='1.0',
    description='RTL simulator API',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email='michal.o.socials@gmail.com',
    install_requires=[
        "jinja2",  # template engine
        "sortedcontainers",  # for calendar queue in simulator
        "pyMathBitPrecise>=0.8",  # bit precise integer types for sim
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "Topic :: System :: Hardware",
        "Topic :: System :: Emulators",
        "Topic :: Utilities"],
    license='MIT',
    packages=find_packages(),
    package_data={'pycocotb.verilator': ['*.h', '*.cpp', '*.template']},
    include_package_data=True,
    zip_safe=False,
    ext_modules=ext_modules,
    test_suite="pycocotb.tests.all.suite"
)



from inspect import isgeneratorfunction

from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.triggers import Edge, WaitCombRead


class CallbackLoop(object):
    """
    Simple utility: process which only register other process/function
    as on change callback for specified signal as soon as it is executed.
    """

    def __init__(self, sim: HdlSimulator, sig: "RtlSignal", fn, shouldBeEnabledFn):
        """
        :param sig: signal on which write callback should be used
        :attention: if condFn is None callback function is always executed

        :ivra fn: function/generator which is callback which should be executed
        :ivar ~.isGenerator: flag if callback function is generator
            or normal function
        :ivar ~.shouldBeEnabledFn: function() -> bool, which returns True if this
            callback loop should be enabled
        :ivar ~.pre_init: if True the 'fn' is executed once imidiately for an intialization
            before any callback is triggered
        """
        assert not isinstance(fn, CallbackLoop)
        self.sim = sim
        self.fn = fn
        self.isGenerator = isgeneratorfunction(fn)
        self.shouldBeEnabledFn = shouldBeEnabledFn
        self._callbackIndex = None
        self._enable = True
        self.sig = sig
        self.pre_init = False

    def setEnable(self, en):
        self._enable = en

    def __call__(self):
        """
        Process for injecting of this callback loop into simulator
        """
        if self.pre_init:
            yield from self.fn()

        while True:
            yield Edge(self.sig)
            if self._enable and self.shouldBeEnabledFn():
                if self.isGenerator:
                    yield from self.fn()
                else:
                    self.fn()


class OnRisingCallbackLoop(CallbackLoop):
    """
    Simple utility: process which only register other process/function
    as on rising callback for specified signal as soon as it is executed.
    """

    def __call__(self):
        if self.pre_init:
            yield from self.fn()

        while True:
            yield Edge(self.sig)
            if self._enable and self.shouldBeEnabledFn():
                yield WaitCombRead()
                if int(self.sig.read()) == 1:
                    if self.isGenerator:
                        yield from self.fn()
                    else:
                        self.fn()


class OnFallingCallbackLoop(CallbackLoop):
    """
    Simple utility: process which only register other process/function
    as on falling callback for specified signal as soon as it is executed.
    """

    def __call__(self):
        if self.pre_init:
            yield from self.fn()

        while True:
            yield Edge(self.sig)
            if self._enable and self.shouldBeEnabledFn():
                yield WaitCombRead()
                if int(self.sig.read()) == 0:
                    if self.isGenerator:
                        yield from self.fn()
                    else:
                        self.fn()




from enum import Enum
from typing import Tuple, Callable, Generator, Optional

from pycocotb.basic_hdl_simulator.io import BasicRtlSimIo
from pycocotb.basic_hdl_simulator.model import BasicRtlSimModel
from pycocotb.basic_hdl_simulator.proxy import BasicRtlSimProxy
from pycocotb.basic_hdl_simulator.sim_utils import ValueUpdater,\
    ArrayValueUpdater


class BasicRtlSimulatorSt(Enum):
    PRE_SET = 0
    EVAL_COMB = 1  # eval circuit without update of memory elements
    EVAL_SEQ = 2  # update whole circuit


def isEvDependentOn(sig: BasicRtlSimProxy, process) -> bool:
    """
    Check if hdl process has event dependency on signal
    """
    if sig is None:
        return False

    return process in sig.simFallingSensProcs\
        or process in sig.simRisingSensProcs


class BasicRtlSimulator():
    COMB_UPDATE_DONE = 0  # all non edge dependent updates done"},
    BEFORE_EDGE = 1  # before evaluation of edge dependent event"},
    END_OF_STEP = 2  # all parts of circuit updated and stable

    def __init__(self):
        self._init_main()
        self._init_listeners()

    def _init_main(self):
        # container of signals in simulation
        self.io = None  # type: BasicRtlSimIo
        self.model = None  # type: BasicRtlSimModel
        self.time = 0  # actual simulation time
        # if true the IO can be only read if false the IO can be only written
        self.read_only_not_write_only = False
        self.pending_event_list = []  # List of triggered callbacks
        self.state = BasicRtlSimulatorSt.PRE_SET
        self._proc_outputs = {}
        self._updates_to_apply = []
        self.signals_checked_for_change = set()
        self._comb_procs_to_run = set()
        self._seq_procs_to_run = set()
        self._updated_in_this_step = set()
        self.needs_init = True

    def _init_listeners(self):
        self.logChange = False
        self.logPropagation = False
        self.logApplyingValues = False

    def bound_model(self, model: BasicRtlSimModel):
        self.model = model
        self.io = model.io
        self._bound_model_procs(model)

    def _bound_model_procs(self, m: BasicRtlSimModel):
        for p, outputs in m._outputs.items():
            assert p not in self._proc_outputs
            self._proc_outputs[p] = tuple(outputs)
        for u in m._units:
            self._bound_model_procs(u)

    def _init_model_signals(self, model: BasicRtlSimModel) -> None:
        """
        * Inject default values to simulation
        * Instantiate IOs for every process
        """
        # set initial value to all signals and propagate it
        for s in model._interfaces:
            if s.def_val is not None:
                s._apply_update(ValueUpdater(s.def_val, False))

        for u in model._units:
            self._init_model_signals(u)

        for p in model._processes:
            self._add_hdl_proc_to_run(None, p)

    def _add_hdl_proc_to_run(self, trigger: Optional[BasicRtlSimProxy], proc) -> None:
        """
        Add hdl process to execution queue
        :param trigger: instance of SimSignal
        :param proc: python generator function representing HDL process
        """
        # first process in time has to plan executing of apply values on the
        # end of this time
        if isEvDependentOn(trigger, proc):
            if self.time == 0:
                return  # pass event dependent on startup
            self._seq_procs_to_run.add(proc)
        else:
            self._comb_procs_to_run.add(proc)

    def _mkUpdater(self, newValue)\
            -> Tuple[Callable[["Value"], bool], bool]:
        """
        This functions resolves write conflicts for signal
        """
        invalidate = False
        if len(newValue) == 3:
            # update for item in array
            val, indexes, isEvDependent = newValue
            return (ArrayValueUpdater(val, indexes, invalidate), isEvDependent)
        else:
            # update for simple signal
            val, isEvDependent = newValue
            return (ValueUpdater(val, invalidate), isEvDependent)

    def _run_comb_processes(self) -> None:
        """
        Delta step for combinational processes
        """
        while self._comb_procs_to_run:
            for proc in self._comb_procs_to_run:
                io = self._proc_outputs[proc]
                proc()
                for sig in io:
                    if sig.val_next is not None:
                        res = self._mkUpdater(sig.val_next)
                        # prepare update
                        updater, is_event_dependent = res
                        self._updates_to_apply.append(
                            (sig, updater, is_event_dependent, proc)
                        )
                        sig.val_next = None
                        # else value is latched
            self._comb_procs_to_run.clear()

            va = self._updates_to_apply
            # log if there are items to log
            lav = self.logApplyingValues
            if va and lav:
                lav(self, va)
            self._updates_to_apply = []

            # apply values to signals, values can overwrite each other
            # but each signal should be driven by only one process and
            # it should resolve value collision
            addSp = self._seq_procs_to_run.add
            for s, vUpdater, isEventDependent, comesFrom in va:
                if isEventDependent:
                    # now=0 and this was process initialization or async reg
                    addSp(comesFrom)
                else:
                    # regular combinational process
                    s._apply_update(vUpdater)

    def _run_seq_processes(self) -> Generator[None, None, None]:
        """
        Delta step for event dependent processes
        """
        updated = []
        for proc in self._seq_procs_to_run:
            io = self._proc_outputs[proc]
            proc()
            updated.extend(io)

        self._seq_procs_to_run = set()

        for sig in updated:
            if sig.val_next is not None:
                v = self._mkUpdater(sig.val_next)
                updater, _ = v
                sig._apply_update(updater)
                sig.val_next = None

    def eval(self):
        "single simulation step"
        if self.needs_init:
            self._init_model_signals(self.model)
            self.needs_init = False

        st = self.state
        if st == BasicRtlSimulatorSt.PRE_SET:
            # apply all writes from outside world
            self.state = BasicRtlSimulatorSt.EVAL_COMB
            self.read_only_not_write_only = True
            self._run_comb_processes()
            return self.COMB_UPDATE_DONE
        elif st == BasicRtlSimulatorSt.EVAL_COMB:
            # evaluate all combinational paths
            # without updating any memory element
            self.state = BasicRtlSimulatorSt.EVAL_SEQ
            self._run_comb_processes()
            return self.BEFORE_EDGE
        elif st == BasicRtlSimulatorSt.EVAL_SEQ:
            # update rest of the circuit including memories
            self._run_comb_processes()
            self._run_seq_processes()
            self._run_comb_processes()
            self.state = BasicRtlSimulatorSt.PRE_SET
            self._updated_in_this_step.clear()
            return self.END_OF_STEP
        else:
            raise AssertionError("Invalid state", self.state)

    def reset_eval(self):
        """
        reset evaluation in COMB_UPDATE_DONE state
        so the comb. circuits can be evaluated again
        """
        assert self.state == BasicRtlSimulatorSt.EVAL_COMB, (self.state, self.time)
        self.state = BasicRtlSimulatorSt.PRE_SET
        self.read_only_not_write_only = False

    def set_trace_file(self, file_name, trace_depth):
        """
        set file where data from signals should be stored

        :param file_name: name of file where trace should be stored (path of vcd file f.e.)
        :param trace_depth: number of hyerarchy levels which should be trraced (-1 = all)
        """
        raise NotImplementedError()

    def set_write_only(self):
        """
        set simulation to write only state, should be called
        before entering to new evaluation step"""
        self.read_only_not_write_only = False
        assert self.state == BasicRtlSimulatorSt.PRE_SET, self.state

    def finalize(self):
        "flush output and clean all pending actions"



"""
Module which generates, compiles and loads Python C-extension which is wrapper around Verilator simulation.
Inputs are Verilog files and result is Python class of simulator which can be instantiated and simulated.
"""


from typing import List, Union, Tuple

from pycocotb.basic_hdl_simulator.proxy import BasicRtlSimProxy


def sensitivity(
        proc_fn,
        *sensitiveTo: List[Union[BasicRtlSimProxy,
                                 Tuple[Tuple[bool, bool], BasicRtlSimProxy]]]):
    """
    register sensitivity for process
    """
    for s in sensitiveTo:
        if isinstance(s, tuple):
            (rising, falling), s = s
            if rising:
                s.simRisingSensProcs.add(proc_fn)
            if falling:
                s.simFallingSensProcs.add(proc_fn)
            assert rising or falling, s
        else:
            s.simSensProcs.add(proc_fn)


def connectSimPort(sim_unit, sim_sub_unit, unit_port_name: str, sub_unit_port_name: str):
    """
    Connect ports of simulation models by name
    (Replace a child port with a parent signal/port directly)
    """
    newPort = getattr(sim_unit.io, unit_port_name)
    setattr(sim_sub_unit.io, sub_unit_port_name, newPort)



from pycocotb.agents.base import AgentBase
from pycocotb.constants import CLK_PERIOD
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.triggers import Timer, WaitWriteOnly


class PullUpAgent(AgentBase):
    """
    After specified time value of the signal is set to 1
    :note: usually used for negated reset
    """

    def __init__(self, sim: HdlSimulator, intf: "RtlSignal", initDelay=int(0.6 * CLK_PERIOD)):
        super(PullUpAgent, self).__init__(sim, intf)
        assert isinstance(initDelay, int)
        self.initDelay = initDelay
        self.data = []

    def driver(self):
        sig = self.intf
        yield WaitWriteOnly()
        sig.write(0)
        yield Timer(self.initDelay)
        yield WaitWriteOnly()
        sig.write(1)


class PullDownAgent(AgentBase):
    """
    After specified time value of the signal is set to 0
    :note: usually used for reset
    """

    def __init__(self, sim: HdlSimulator, intf: "RtlSignal", initDelay=int(0.6 * CLK_PERIOD)):
        super(PullDownAgent, self).__init__(sim, intf)
        assert isinstance(initDelay, int)
        self.initDelay = initDelay
        self.data = []

    def driver(self):
        sig = self.intf
        yield WaitWriteOnly()
        sig.write(1)
        yield Timer(self.initDelay)
        yield WaitWriteOnly()
        sig.write(0)



from collections import deque
from typing import Tuple, Union

from pycocotb.agents.base import AgentWitReset, NOP
from pycocotb.constants import CLK_PERIOD
from pycocotb.hdlSimulator import HdlSimulator
from pycocotb.simCalendar import DONE
from pycocotb.triggers import Timer, WaitWriteOnly, WaitCombRead, Edge


class TristateSignal():
    """
    Container of signals for controll of tristate block

    :ivar ~.i: input - slave to master
    :ivar ~.o: output - master to slave
    :ivar ~.t: master to slave, if 1 the value of o is set to i
    """
    def __init__(self, i: "RtlSignal", o: "RtlSignal", t: "RtlSignal"):
        self.i = i
        self.o = o
        self.t = t


class TristateAgent(AgentWitReset):
    """
    :ivar ~.selfSynchronization: if True the agent reads/write
        with a perior of DEFAULT_CLOCK
    :ivar ~.pullMode: specifies how the interface should behave if none drives it
        can be (1: pull-up, 0: pull-down, None: disconnected)
    """

    def __init__(self,
                 sim: HdlSimulator,
                 intf: TristateSignal,
                 rst: Tuple["RtlSignal", bool]):
        """
        :param intf: tuple (i signal, o signal, t signal)
            as present in tristate interface
        :note: t signal controls if the output should be connected,
            if 't'=0 the 'o' does not have effect
        """
        super(TristateAgent, self).__init__(sim, intf, rst)
        self.i, self.o, self.t = intf.i, intf.o, intf.t
        self.data = deque()
        # can be (1: pull-up, 0: pull-down, None: disconnected)
        self.pullMode = 1  # type: Union[1, 0, None]
        self.selfSynchronization = True
        self.collectData = True

    def monitor(self):
        """
        The evaluate a tristate 'i' value from 'o' and 't'
        and optionaly store it.
        One step.
        """
        yield WaitCombRead()
        # read in pre-clock-edge
        t = self.t.read()
        o = self.o.read()
        sim = self.sim
        if self.pullMode is not None and sim.now > 0:
            try:
                t = int(t)
            except ValueError:
                raise AssertionError(
                    sim.now, self.t, "Invalid value on tristate interface => ioblock would burn")
            try:
                o = int(o)
            except ValueError:
                raise AssertionError(
                    sim.now, self.o, "Invalid value on tristate interface => ioblock would burn")

            if self.pullMode == o:
                raise AssertionError(
                    sim.now, self.o, "Can not set value to a same as pull up,"
                    " because others may try to set it to oposite => ioblock would burn")

        if t:
            v = o
        else:
            v = self.pullMode

        last = self.i.read()
        try:
            last = int(last)
        except ValueError:
            last = None

        yield WaitWriteOnly()
        self.i.write(v)

        if self.collectData and sim.now > 0:
            yield WaitCombRead()
            if self.notReset():
                self.data.append(v)

    def getMonitors(self):
        return [self.onTWriteCallback(), ]

    def onTWriteCallback(self):
        while True:
            yield Edge(self.t, self.o)
            if self.getEnable():
                # if we are this signal was update by change of some memory we can not write in this
                # time slot and we have to wait for another
                if self.sim._current_time_slot.write_only is DONE:
                    yield Timer(1)
                yield from self.monitor()

    def _write(self, val):
        """
        Update value on interface.

        :type val: Union[int, NOP]
        """
        if val is NOP:
            # control now has slave
            t = 0
            o = self.pullMode
        else:
            # control now has this agent
            t = 1
            o = val

        self.t.write(t)
        self.o.write(o)

    def _read(self):
        """
        :return: actual value on interface
        """
        return self.i.read()

    def driver(self):
        """
        Drive 'o' and 't' from data buffer.
        One step if not selfSynchronization else infinite loop.
        """
        while True:
            yield WaitWriteOnly()
            if self.data:
                o = self.data.popleft()
                if o == NOP:
                    t = 0
                    o = 0
                else:
                    t = 1
                self.o.write(o)
                self.t.write(t)

            if self.selfSynchronization:
                yield Timer(CLK_PERIOD)
            else:
                break


class TristateClkAgent(TristateAgent):
    """
    Agent for tri-state interface which generates clock signal
    and ignores all other components which are trying to drive this clk signal
    """

    def __init__(self, sim: HdlSimulator, intf, rst: Tuple["RtlSignal", bool]):
        super(TristateClkAgent, self).__init__(sim, intf, rst)
        self.period = CLK_PERIOD
        self.collectData = False

    def driver(self):
        o = self.o
        high = self.pullMode
        low = not self.pullMode
        halfPeriod = self.period // 2

        yield WaitWriteOnly()
        o.write(low)
        self.t.write(1)

        while True:
            yield Timer(halfPeriod)
            yield WaitWriteOnly()
            o.write(high)

            yield Timer(halfPeriod)
            yield WaitWriteOnly()
            o.write(low)
